package com.flutter_webview_plugin;

import android.annotation.TargetApi;
import android.graphics.Bitmap;
import android.os.Build;
import android.util.Log;
import android.webkit.WebResourceRequest;
import android.webkit.WebResourceResponse;
import android.webkit.WebView;
import android.webkit.WebViewClient;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Created by lejard_h on 20/12/2017.
 */

public class BrowserClient extends WebViewClient {
    private static final String TAG = BrowserClient.class.getSimpleName();

    private Pattern invalidUrlPattern = null;
    List<String> overridableUrls;
    List<String> uriSchemes;
    String currentUrl;
    Map<String,String> allRequestHeaders;

    public BrowserClient() {
        this(null);
    }

    public BrowserClient(String invalidUrlRegex) {
        super();
        if (invalidUrlRegex != null) {
            invalidUrlPattern = Pattern.compile(invalidUrlRegex);
        }
    }

    public void updateInvalidUrlRegex(String invalidUrlRegex) {
        if (invalidUrlRegex != null) {
            invalidUrlPattern = Pattern.compile(invalidUrlRegex);
        } else {
            invalidUrlPattern = null;
        }
    }

    @Override
    public void onPageStarted(WebView view, String url, Bitmap favicon) {
        super.onPageStarted(view, url, favicon);
        Log.d(TAG, "onPageStarted: " + url);
        Map<String, Object> data = new HashMap<>();
        data.put("url", url);
        data.put("type", "startLoad");
        FlutterWebviewPlugin.channel.invokeMethod("onState", data);
    }

    @TargetApi(Build.VERSION_CODES.LOLLIPOP)
    @Override
    public boolean shouldOverrideUrlLoading(WebView view, WebResourceRequest request) {
        // returning true causes the current WebView to abort loading the URL,
        // while returning false causes the WebView to continue loading the URL as usual.

        String url = request.getUrl().toString();
        return handleUrl(view, url);
    }

    @Override
    public boolean shouldOverrideUrlLoading(WebView view, String url) {
        // returning true causes the current WebView to abort loading the URL,
        // while returning false causes the WebView to continue loading the URL as usual.
        return handleUrl(view, url);
    }

    private boolean handleUrl(WebView view, String url) {
        Log.d(TAG, "shouldOverrideUrlLoading: " + url);
        if (schemeMatched(url)) {
            Log.d(TAG, "schemeMatched: " + url);

            Map<String, Object> data = new HashMap<>();
            data.put("url", url);
            FlutterWebviewPlugin.channel.invokeMethod("onSchemeMatched", data);
            return true;
        } else if (canUrlBeOverridden(url)) {
            Log.d(TAG, "canUrlBeOverridden: " + url);

            Map<String, Object> data = new HashMap<>();
            data.put("url", url);
            FlutterWebviewPlugin.channel.invokeMethod("onUrlOverridden", data);
            return true;
        } else {
            boolean isInvalid = checkInvalidUrl(url);

            Log.d(TAG, url + " is " + (isInvalid ? "invalid!" : "valid!"));

            Map<String, Object> data = new HashMap<>();
            data.put("url", url);
            data.put("type", isInvalid ? "abortLoad" : "shouldStart");

            FlutterWebviewPlugin.channel.invokeMethod("onState", data);

            if(!isInvalid && (currentUrl == null || !url.contains(currentUrl)) && allRequestHeaders!=null){
                Log.d(TAG, "load with allRequestHeaders.");
                currentUrl = url;
                view.loadUrl(url, allRequestHeaders);
                return true;
            }

            return isInvalid;
        }
    }

    @Override
    public WebResourceResponse shouldInterceptRequest(WebView view, WebResourceRequest request) {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
            Log.d(TAG, "shouldInterceptRequest: allRequestHeaders= "+request.getRequestHeaders());
        }
        return super.shouldInterceptRequest(view, request);
    }

    @TargetApi(Build.VERSION_CODES.LOLLIPOP)
    @Override
    public void onReceivedHttpError(WebView view, WebResourceRequest request, WebResourceResponse errorResponse) {
        super.onReceivedHttpError(view, request, errorResponse);
        handleReceivedError(view, errorResponse.getStatusCode(), errorResponse.getReasonPhrase(), request.getUrl().toString());
    }

    @Override
    public void onReceivedError(WebView view, int errorCode, String description, String failingUrl) {
        super.onReceivedError(view, errorCode, description, failingUrl);
        handleReceivedError(view, errorCode, description, failingUrl);
    }

    private void handleReceivedError(WebView view, int errorCode, String description, String failingUrl) {
        Log.d(TAG, String.format("handleReceivedError: url=%s, description:%s, errorCode=%s", failingUrl, description, errorCode));

        Map<String, Object> data = new HashMap<>();
        data.put("url", failingUrl);
        data.put("code", errorCode);
        FlutterWebviewPlugin.channel.invokeMethod("onHttpError", data);
    }

    @Override
    public void onPageFinished(WebView view, String url) {
        super.onPageFinished(view, url);
        Log.d(TAG, "onPageFinished: " + url);

        Map<String, Object> data = new HashMap<>();
        data.put("url", url);

        FlutterWebviewPlugin.channel.invokeMethod("onUrlChanged", data);

        data.put("type", "finishLoad");
        FlutterWebviewPlugin.channel.invokeMethod("onState", data);
    }

    private boolean checkInvalidUrl(String url) {
        if (invalidUrlPattern == null) {
            return false;
        } else {
            Matcher matcher = invalidUrlPattern.matcher(url);
            return matcher.lookingAt();
        }
    }

    public void setOverridableUrls(List<String> overridableUrls) {
        this.overridableUrls = overridableUrls;
    }

    public void setUriSchemes(List<String> uriSchemes) {
        this.uriSchemes = uriSchemes;
    }

    private boolean schemeMatched(String url) {
        if (uriSchemes != null) {
            for (String scheme : uriSchemes) {
                if (url.startsWith(scheme)) {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean canUrlBeOverridden(String url) {
        if (overridableUrls != null) {
            for (String oUrl : overridableUrls) {
                if (url.contains(oUrl)) {
                    return true;
                }
            }
        }
        return false;
    }

    public void setCurrentUrl(String currentUrl) {
        this.currentUrl = currentUrl;
    }

    public void setAllRequestHeaders(Map<String, String> allRequestHeaders) {
        this.allRequestHeaders = allRequestHeaders;
    }
}